import java.awt.*;
import java.applet.*;
import java.awt.event.*;
import java.util.*;
import java.net.*;

/*****************************************************************
* An Applet built to illustrate the use of the Projectile class
* that is assignment #9 of CS1E in 1999.
* Fundamentally the applet draws a random, rough, mountainous
* 2D terrain. Two TextFields allow the input of an angle (degrees)
* and muzzle velocity (metres/sec) with which a mortar shell
* (Projectile) is fired. A Projectile object is constructed based on
* the inputs and used to animate the location of a shell on a
* 5ms basis.
* @author Spike Barlow
* @date 27/6/199
*******************************************************************/
public class Mortar extends Applet
                implements ActionListener {

  boolean application = false;

  /**  Mortar shell */
  Projectile            shell;

  /** 3 Constants that define what state the applet is currently in. */
  static final int      WAITING = 0;
  static final int      IN_FLIGHT = 1;
  static final int      EXPLODING = 2;
  int                   state = WAITING;

  /** Elapsed flight time for the shell. */
  double                time = 0;

  /** The terrain. Another class. STEEPNESS is a parameter passed
  * to the constructor to indicate how rough the terrain is */
  RandomTerrain         landscape;
  static final double   STEEPNESS = 4;

  /** The various GUI components needed to allow user input
  * and the right appearance */
  TextField             angleField;
  TextField             muzzleField;
  Button                fireButton;
  Panel                 controlPanel;

  /** Various sizes in pixels. These should really be derived from
  * the html file with Applet's getSize() call. */
  static final int      PLOTXSIZE = 600;
  static final int      PLOTYSIZE = 500;
  static final int      CONTROLXSIZE = 150;
  static final int      CONTROLYSIZE = 400;

  /** The time-step in animating the shell. Time is advanced by this
  * amount and the shell's new location determined. Too large a value
  * and the shell may pass through objects and jerk across the screen.
  * Too small and the animation takes forever. */
  static final double   TIMESTEP = 0.05;

  /** Current co-ordinates of the shell */
  double                currentX,
                        currentY;

  /** The scale factors for mapping between user space and the
  * dimensions (pixels) of the screen. A good implementation would
  * determine these on the basis of the size of the terrain coupled
  * with the Applet size. Also an initial pixel height at which the
  * terrain is drawn, the size of the shell in pixels and some
  * important colours. */
  static final double   XSCALE = 1.0;
  static final double   YSCALE = 1.0;
  static final int      STARTHEIGHT = 300;
  static final int      SHELLSIZE = 6;
  Color    backgroundColour = Color.white;
  static final Color    landColour = new Color(0.8f,0.7f,0.7f);

  /** Burst data for the mortar shells when they impact the landscape */
  static final double   BURSTRADIUS = 10.0;
  double[]   craterDepths;
  static final double   MAX_MUZZLE = 100.0;

  /** Cannon co-ordinates. */
  int           cannonX;
  double        cannonY;

  /** Constants used in ensuring the cannon always starts at a random,
  * but reasonable location on the landscape. */
  static final int      MIN_CANNON_X = 5;
  static final int      CANNON_X_RANGE = 20;
  static final int      CANNON_PLATEAU = 40;

  /** Sounds. */
//  AudioClip     explosionSound;


/*******************************************************************
* Initialise the applet. Generate the terrain. Find and flatten a
* starting location for the cannon. Add the appropriate GUI components
* and listen for a button push which indicates a firing of the mortar.
**********************************************************************/
public void init() {
  int           i;

  Random        generator = new Random();

/* Commented out code to play sounds because it causes jerky,
   sporadic animation despite the fact that the sound playing
   was run on a separate thread. This may simply be an audio
   set-up problem specific to my PC (on which I tested it) but...
  if (application) {
    try {explosionSound = getAudioClip(new URL("Explode.wav"));}
    catch (Exception e) {}
  }
  else
    explosionSound = getAudioClip(getCodeBase(),"Explode.wav");
*/
  craterDepths = new double[(int)Math.round(BURSTRADIUS)];
  for (i=0;i<craterDepths.length;i++)
    craterDepths[i] = Math.sqrt(BURSTRADIUS*BURSTRADIUS-i*i);
  landscape = new RandomTerrain(0,PLOTXSIZE,STEEPNESS,-200,250);
  cannonX = MIN_CANNON_X + Math.abs(generator.nextInt()) % CANNON_X_RANGE;
  cannonY = landscape.getHeightAt(cannonX)+SHELLSIZE+5.0;
  landscape.flatten(cannonX,CANNON_PLATEAU);
  controlPanel = new Panel();
  controlPanel.setSize(CONTROLXSIZE,CONTROLYSIZE);
  add(controlPanel);
  controlPanel.add(new Label("Angle of Fire",Label.RIGHT));
  angleField = new TextField(10);
  controlPanel.add(angleField);
  controlPanel.add(new Label("Muzzle Velocity",Label.RIGHT));
  muzzleField = new TextField(10);
  controlPanel.add(muzzleField);
  fireButton = new Button("Fire");
  controlPanel.add(fireButton);
  fireButton.addActionListener(this);
}

/*******************************************************************
* Paint the Applet to the screen. Draws the terrain using its
* drawTerrain method. If there is a shell in flight then the a loop
* is used to calculate the shell location at TIMESTEP intervals
* until it hits the ground or passes off the edge of the Applet
* (left or right edge). This is NOT the best way to write an Applet
* as the loop ties up the browser. For "normal" shots that is fine
* but for high velocity shots straight up (90 degrees) that could
* be a long time. Ideally either update should be over-ridden to
* eliminate the screen clearing (and hence flicker) and repaint
* called. Better yet a Thread should be started to handle the
* Projectile calculations. Any of these approaches though would
* further obscure a program who's prime purpose is teaching at 1st
* year, 5 weeks into the course, level.
*********************************************************************/
public void paint(Graphics g) {
  double        previousX,
                previousY;

  g.setColor(backgroundColour);
  g.fillRect(0,0,PLOTXSIZE,PLOTYSIZE);
  g.setColor(Color.black);
  landscape.drawTerrain(g,1,STARTHEIGHT,XSCALE,YSCALE,landColour,
        PLOTXSIZE,PLOTYSIZE);
  if (state==IN_FLIGHT) {
    previousX = shell.getHorizontalPosition(0);
    previousY = shell.getVerticalPosition(0);
    currentX = shell.getHorizontalPosition(time);
    currentY = shell.getVerticalPosition(time);
    // Keep looping while the shell is in flight. 
    while (currentX>=0 && currentX<=PLOTXSIZE) {
      double terrainHeight = landscape.getHeightAt((int)Math.round(currentX));
      if (currentY-SHELLSIZE>terrainHeight) {
        g.setColor(backgroundColour);
        // Fill in the old location of the shell with the background
        // colour. Far more efficient than redrawing the entire screen.
        g.fillOval((int)Math.round(previousX*XSCALE),
                STARTHEIGHT-(int)Math.round(previousY*YSCALE),
                SHELLSIZE,SHELLSIZE);
        // Draw the shell at its current location.
        g.setColor(Color.black);
        g.fillOval((int)Math.round(currentX*XSCALE),
                STARTHEIGHT-(int)Math.round(currentY*YSCALE),
                SHELLSIZE,SHELLSIZE);
        time += TIMESTEP;
        try { Thread.sleep(10); }
        catch (Exception e) { }
        previousX = currentX;
        previousY = currentY;
        currentX = shell.getHorizontalPosition(time);
        currentY = shell.getVerticalPosition(time);
      }
      else {
      // Shell has hit the ground. For best effect should animate an
      // explosion and play an appropriate sound.
      // SoundPlayingThread player = new SoundPlayingThread(explosionSound);
      // player.start();
        g.drawString("Bang!",
                (int)Math.round(currentX),
                STARTHEIGHT-(int)Math.round(terrainHeight));
        landscape.addCrater((int)Math.round(currentX),craterDepths);
        break;
      }
    }
  state = WAITING;
  time = 0;
  repaint();
  }       
}

/*********************************************************************
* An action has been performed on one of the GUI components. As the
* Button is the only thing being listened to there is no need to check
* which one. Grab the text in each of the TextFields ad attempt to
* turn it into a real (double) value. Start up a shell (Projectile)
* and call (indirectly through repaint()) the paint method.
**********************************************************************/
public void actionPerformed(ActionEvent e) {
  double        angle = 0;
  double        muzzle = 0;

  try {
    angle = Double.valueOf(angleField.getText()).doubleValue();
    muzzle = Double.valueOf(muzzleField.getText()).doubleValue();
    muzzle = Math.min(muzzle,MAX_MUZZLE);
    muzzleField.setText(""+muzzle);
    shell = new Projectile(cannonX,cannonY,angle,muzzle);
    state = IN_FLIGHT;
    time = TIMESTEP;
    repaint();
  }
  catch (NumberFormatException nfe) {state=WAITING; }
} 


/*******************************************************************
* Allow the Applet to run as a stand-alone application also. Creates
* a Frame, creates an instance of the Applet and adds it to the Frame.
* Explicitly calls the init() and start() methods of the Applet. Note
* that it does not listen for WindowClosing events (i.e., need to
* cntrl-c out of the program).
*********************************************************************/
public static void main(String[] args) {

  Frame f = new Frame("Mortar Applet");
  f.setSize(PLOTXSIZE,PLOTYSIZE);
  f.setLayout(new BorderLayout());
  Mortar mortarApplet = new Mortar();
  mortarApplet.application = true;
  f.add(mortarApplet,"Center");
  mortarApplet.init();
  mortarApplet.start();
  f.setVisible(true);
}

}
